import logging
from logging import LogRecord
from typing import Dict

import bugsnag
from bugsnag.breadcrumbs import BreadcrumbType

__all__ = ('BugsnagHandler',)


class BugsnagHandler(logging.Handler, object):
    def __init__(self, client=None, extra_fields=None):
        """
        Creates a new handler which sends records to Bugsnag
        """
        super(BugsnagHandler, self).__init__()
        self.client = client
        self.custom_metadata_fields = extra_fields
        self.callbacks = [self.extract_default_metadata,
                          self.extract_custom_metadata,
                          self.extract_severity]

    def emit(self, record: LogRecord):
        """
        Outputs the record to Bugsnag
        """
        if hasattr(bugsnag, '__path__'):
            paths = getattr(bugsnag, '__path__')
            for path in paths:
                if path in record.pathname:
                    return

        options = {
            'metadata': {},
            'unhandled': False,
            'severity_reason': {
                'type': 'log',
                'attributes': {
                    'level': record.levelname
                }
            }
        }

        client = self.client or bugsnag.legacy.default_client

        for callback in self.callbacks:
            try:
                callback(record, options)
            except Exception as e:
                client.configuration.logger.error(
                    'Failed to run handler callback %s',
                    e
                )

        if 'exception' in options:
            exception = options.pop('exception')
            if isinstance(exception, Exception):
                client.notify(exception, **options)
                return
            else:
                try:
                    metadata = options['metadata']
                    if isinstance(metadata, dict):
                        metadata['exception'] = exception
                except TypeError:
                    # Means options['metadata'] is no longer a dictionary
                    pass

        if record.exc_info:
            client.notify_exc_info(*record.exc_info, **options)
        else:
            # Create exception type dynamically, to prevent bugsnag.handlers
            # being prepended to the exception name due to class name
            # detection in utils. Because we are messing with the module
            # internals, we don't really want to expose this class anywhere
            level_name = record.levelname or "Message"
            exc_type = type('Log' + level_name, (Exception, ), {})
            exc = exc_type(record.getMessage())
            exc.__module__ = None

            client.notify(exc, **options)

    def add_callback(self, callback):
        """
        Add a new callback to be invoked after an log message is sent but
        before it is sent to Bugsnag. Each callback is invoked with the
        LogRecord and options to be sent to notify.
        """
        self.callbacks.append(callback)

    def remove_callback(self, callback):
        """
        Remove a callback
        """
        self.callbacks.remove(callback)

    def clear_callbacks(self):
        """
        Clear all callbacks
        """
        del self.callbacks[:]

    def extract_severity(self, record: LogRecord, options: Dict):
        """
        Convert log record level into severity levels
        """
        levelno = record.levelno or logging.WARNING

        if levelno >= logging.ERROR:
            options['severity'] = 'error'
        elif levelno >= logging.WARNING:
            options['severity'] = 'warning'
        else:
            options['severity'] = 'info'

    def extract_custom_metadata(self, record: LogRecord, options: Dict):
        """
        Append the contents of selected fields of a record to the metadata
        of a report
        """
        if self.custom_metadata_fields is None:
            return

        if 'metadata' not in options:
            options['metadata'] = {}

        for section in self.custom_metadata_fields:
            if section not in options['metadata']:
                options['metadata'][section] = {}

            for field in self.custom_metadata_fields[section]:
                if hasattr(record, field):
                    attr = getattr(record, field)
                    options['metadata'][section][field] = attr

    def extract_default_metadata(self, record: LogRecord, options: Dict):
        """
        Extract log record fields into error report metadata
        """
        record_fields = ('asctime', 'created', 'levelname', 'levelno', 'msecs',
                         'name', 'process', 'processName', 'relativeCreated',
                         'thread', 'threadName',)

        extra_data = {}
        for field in record_fields:
            if hasattr(record, field):
                extra_data[field] = getattr(record, field)

        if 'metadata' not in options:
            options['metadata'] = {}

        options['metadata']['extra data'] = extra_data

    def leave_breadcrumbs(self, record):
        """
        A log filter that leaves breadcrumbs for log records with a level above
        or equal to "configuration.breadcrumb_log_level" and below the
        handler's level
        """

        client = self.client or bugsnag.legacy.default_client

        # Only leave a breadcrumb if we aren't going to notify this log record
        # and its "bugsnag_create_breadcrumb" attribute isn't False
        # If the handler has no level (is NOTSET) then we will leave a
        # breadcrumb because it's likely this means the handler has not been
        # attached. Otherwise it would be notify-ing for every log call
        if (
            record.levelno >= client.configuration.breadcrumb_log_level
            and (record.levelno < self.level or self.level == logging.NOTSET)
            and getattr(record, 'bugsnag_create_breadcrumb', True)
        ):
            client._auto_leave_breadcrumb(
                record.getMessage(),
                {"logLevel": record.levelname},
                BreadcrumbType.LOG
            )

        return True



import flask
from typing import Dict

import bugsnag
from bugsnag.wsgi import request_path
from bugsnag.legacy import _auto_leave_breadcrumb
from bugsnag.breadcrumbs import BreadcrumbType
from bugsnag.utils import remove_query_from_url


__all__ = ('handle_exceptions',)


def add_flask_request_to_notification(event: bugsnag.Event):
    if not flask.request:
        return

    event.request = flask.request
    if event.context is None:
        event.context = "%s %s" % (flask.request.method,
                                   request_path(flask.request.environ))

    if "id" not in event.user:
        event.set_user(id=flask.request.remote_addr)
    event.add_tab("session", dict(flask.session))
    if bugsnag.configure().send_environment:
        event.add_tab("environment", dict(flask.request.environ))
    event.add_tab("request", {
        "url": flask.request.base_url,
        "headers": dict(flask.request.headers),
        "params": dict(flask.request.form),
        "data":
            flask.request.get_json(silent=True) or
            dict(body=flask.request.data)
    })


def handle_exceptions(app):
    middleware = bugsnag.configure().internal_middleware
    bugsnag.configure().runtime_versions['flask'] = flask.__version__
    middleware.before_notify(add_flask_request_to_notification)
    flask.got_request_exception.connect(__log_exception, app)
    flask.request_started.connect(_on_request_started, app)


# pylint: disable-msg=W0613
def __log_exception(sender, exception, **extra):
    bugsnag.auto_notify(exception, severity_reason={
        "type": "unhandledExceptionMiddleware",
        "attributes": {
            "framework": "Flask"
        }
    })


def _on_request_started(sender, **extra):
    if bugsnag.configuration.auto_capture_sessions:
        bugsnag.start_session()

    if flask.request:
        _auto_leave_breadcrumb(
            'http request',
            _get_breadcrumb_metadata(flask.request),
            BreadcrumbType.NAVIGATION
        )


def _get_breadcrumb_metadata(request) -> Dict[str, str]:
    metadata = {'to': request_path(request.environ)}

    if 'referer' in request.headers:
        metadata['from'] = remove_query_from_url(request.headers['referer'])

    return metadata



from typing import Callable, Optional, Type, List

from bugsnag.event import Event


Middleware = Callable[[Event], Callable]


__all__ = []  # type: List[str]


class SimpleMiddleware:
    def __init__(self, before: Optional[Middleware] = None,
                 after: Optional[Middleware] = None):
        self.before = before
        self.after = after

    def __call__(self, bugsnag):

        def middleware(event):
            if self.before:
                ret = self.before(event)
                if ret is False:
                    return

            bugsnag(event)

            if self.after:
                self.after(event)

        return middleware


class DefaultMiddleware:
    """
    DefaultMiddleware provides the transformation from request_config into
    metadata that has always been supported by bugsnag-python.
    """
    def __init__(self, bugsnag: Middleware):
        self.bugsnag = bugsnag

    def __call__(self, event: Event):
        config = event.request_config
        event.set_user(id=config.user_id)
        event.set_user(**config.user)

        if not event.context:
            event.context = config.get("context")

        for name, dictionary in config.metadata.items():
            if name in event.metadata:
                for key, value in dictionary.items():
                    if key not in event.metadata[name]:
                        event.metadata[name][key] = value
            else:
                event.add_tab(name, dictionary)

        event.add_tab("request", config.get("request_data"))
        if event.config.send_environment:
            event.add_tab("environment", config.get("environment_data"))
        event.add_tab("session", config.get("session_data"))
        event.add_tab("extraData", config.get("extra_data"))

        self.bugsnag(event)


def skip_bugsnag_middleware(event: Event):
    """
    A callback-based middleware that prevents notifying an event where the
    'original_error' has a 'skip_bugsnag' attr set to 'True'.
    """
    if getattr(event.original_error, 'skip_bugsnag', False) is True:
        return False


class MiddlewareStack:
    """
    Manages a stack of Bugsnag middleware.
    """
    def __init__(self):
        self.stack = []

    def before_notify(self, func: Middleware):
        """
        Add a function to be run before bugsnag is notified.

        This lets you modify the payload that will be sent.
        If your function returns False, nothing will be sent.

        >>> def add_request_data(event):
        ...    event.add_tab("request", request_data)
        >>>
        >>> stack = MiddlewareStack()
        >>> stack.before_notify(add_request_data)
        """
        self.append(SimpleMiddleware(before=func))

    def after_notify(self, func: Middleware):
        """
        Add a function to be run after bugsnag is notified.

        This lets you log errors in custom ways.
        """
        self.append(SimpleMiddleware(after=func))

    def append(self, middleware: Middleware):
        """
        Add a middleware to the end of the stack.

        It will be run after all middleware currently defined.
        If you want to stop the event progress, return from
        your __call__ method without calling the next level.

        >>> class ExampleMiddleware():
        ...     def __init__(self, bugsnag):
        ...         self.bugsnag = bugsnag
        ...     def __call__(self, event):
        ...         config = event.request_config
        ...         event.add_tab("request", config.get("request"))
        ...         self.bugsnag(event)
        ...
        >>> stack = MiddlewareStack()
        >>> stack.append(ExampleMiddleware)
        """
        self.stack.append(middleware)

    def insert_before(self, target_class: Type, middleware: Middleware):
        """
        Adds a middleware to the stack in the position before
        the target_class.
        """
        try:
            index = self.stack.index(target_class)
            self.stack.insert(index, middleware)
        except ValueError:
            self.append(middleware)

    def insert_after(self, target_class: Type, middleware: Middleware):
        """
        Adds a middleware to the stack in the position after
        the target_class.
        """
        try:
            index = self.stack.index(target_class)
            self.stack.insert(index + 1, middleware)
        except ValueError:
            self.append(middleware)

    def run(self, event: Event, callback: Callable[[], None]):
        """
        Run all the middleware in order, then call the callback.
        """

        # the last step in the event stack is to call the callback.
        # we also do this inside the exception handler, so need to ensure that
        # the callback is only called once.
        def finish(event):
            if not hasattr(finish, 'called'):
                finish.called = True
                callback()

        to_call = finish
        for middleware in reversed(self.stack):
            to_call = middleware(to_call)

        try:
            to_call(event)
        except Exception:
            event.config.logger.exception(
                'Error in exception middleware'
            )

            # still notify if middleware crashes before event
            finish(event)



from enum import Enum, unique
from typing import Any, Dict, List, Optional, Union, Callable, TYPE_CHECKING
from collections import deque

from bugsnag.utils import FilterDict

# Deque is not present in 'typing' until 3.5.4, so we can't use it directly
if TYPE_CHECKING:
    from typing import Deque

# The _breadcrumbs context var contains None or a deque of Breadcrumb instances
try:
    from contextvars import ContextVar
    _breadcrumbs = ContextVar(
        'bugsnag-breadcrumbs',
        default=None
    )  # type: ContextVar[Optional[Deque[Breadcrumb]]]
except ImportError:
    from bugsnag.utils import ThreadContextVar
    _breadcrumbs = ThreadContextVar('bugsnag-breadcrumbs', default=None)  # type: ignore  # noqa: E501


__all__ = (
    'BreadcrumbType',
    'Breadcrumb',
    'Breadcrumbs',
    'OnBreadcrumbCallback'
)


@unique
class BreadcrumbType(Enum):
    NAVIGATION = 'navigation'
    REQUEST = 'request'
    PROCESS = 'process'
    LOG = 'log'
    USER = 'user'
    STATE = 'state'
    ERROR = 'error'
    MANUAL = 'manual'


class Breadcrumb:
    def __init__(
        self,
        message: str,
        type: BreadcrumbType,
        metadata: Dict[str, Any],
        timestamp: str
    ):
        self.message = message
        self.type = type
        self.metadata = metadata
        self._timestamp = timestamp

    @property
    def timestamp(self) -> str:
        return self._timestamp

    # Convert this breadcrumb into a dict for use when JSON encoding
    def to_dict(self) -> Dict[str, Union[str, FilterDict]]:
        return {
            'timestamp': self._timestamp,
            'name': self.message,
            'type': self.type.value,
            'metaData': FilterDict(self.metadata)
        }


class Breadcrumbs:
    def __init__(self, max_breadcrumbs: int):
        self._max_breadcrumbs = max_breadcrumbs

        # Calling resize is important for tests as we make many Breadcrumbs
        # instances but they all have to share a ContextVar, so the size can
        # leak between tests
        self.resize(max_breadcrumbs)

    def append(self, breadcrumb: Breadcrumb) -> None:
        self._breadcrumbs.append(breadcrumb)

    # Resize the list of breadcrumbs if configuration.max_breadcrumbs changes
    def resize(self, new_max: int) -> None:
        old_breadcrumbs = self._breadcrumbs
        new_breadcrumbs = deque(old_breadcrumbs, maxlen=new_max)

        _breadcrumbs.set(new_breadcrumbs)
        self._max_breadcrumbs = new_max

    # Create a copy of the current list of breadcrumbs for this context
    def create_copy_for_context(self) -> None:
        # Resizing will create a new deque and store it in the ContextVar,
        # which will give the current context a new copy of the list
        self.resize(self._max_breadcrumbs)

    def clear(self) -> None:
        self._breadcrumbs.clear()

    def to_list(self) -> List[Breadcrumb]:
        return list(self._breadcrumbs)

    @property
    def _breadcrumbs(self):
        # type: () -> Deque[Breadcrumb]
        try:
            breadcrumbs = _breadcrumbs.get()
        except LookupError:
            breadcrumbs = None

        if breadcrumbs is None:
            breadcrumbs = deque(maxlen=self._max_breadcrumbs)
            _breadcrumbs.set(breadcrumbs)

        return breadcrumbs


OnBreadcrumbCallback = Callable[[Breadcrumb], Union[None, bool]]



from typing import Any, Dict, Optional, List
import linecache
import logging
import os
import sys
import traceback
import inspect
import warnings
from copy import deepcopy

import bugsnag

from bugsnag.breadcrumbs import Breadcrumb
from bugsnag.utils import fully_qualified_class_name as class_name
from bugsnag.utils import FilterDict, package_version, SanitizingJSONEncoder
from bugsnag.error import Error

__all__ = ('Event',)


class Event:
    """
    An occurrence of an exception for delivery to Bugsnag
    """
    NOTIFIER_NAME = "Python Bugsnag Notifier"
    NOTIFIER_URL = "https://github.com/bugsnag/bugsnag-python"
    PAYLOAD_VERSION = "4.0"
    SUPPORTED_SEVERITIES = ["info", "warning", "error"]

    def __init__(self, exception: BaseException, config, request_config,
                 **options):
        """
        Create a new event

        exception is the exception being reported.
        config is the global instance of bugsnag.Configuration
        request_config is the thread-local instance of bugsnag.Configuration
        (used by middleware)

        options can be used to override any of the configuration parameters:
            "api_key", "release_stage", "app_version", "hostname"
        and to provide the following top-level event payload keys:
            "user", "context", "severity", "grouping_hash", "metadata",
            ("user_id")
        or to provide the exception parameter:
            "traceback"
        All other keys will be sent as metadata to Bugsnag.
        """
        self._exception = exception
        self._original_error = exception

        self.options = options
        self.config = config
        self.request_config = request_config
        self.request = None  # type: Any
        self._breadcrumbs = [
            deepcopy(breadcrumb) for breadcrumb in config.breadcrumbs
        ]

        def get_config(key):
            return options.pop(key, getattr(self.config, key))

        self.release_stage = get_config("release_stage")
        self.app_version = get_config("app_version")
        self.app_type = get_config("app_type")
        self.hostname = get_config("hostname")
        self.runtime_versions = get_config("runtime_versions")
        self.send_code = get_config("send_code")

        self.context = options.pop("context", None)
        self.severity = options.pop("severity", "warning")
        if self.severity not in self.SUPPORTED_SEVERITIES:
            self.severity = "warning"

        self.unhandled = options.pop("unhandled", False)
        self.severity_reason = options.pop(
            "severity_reason",
            {'type': 'handledException'}
        )

        self.user = options.pop("user", {})
        if "user_id" in options:
            self.user["id"] = options.pop("user_id")

        # for backwards compatibility we generate the first error's stacktrace
        # here and use it as 'self.stacktrace' and 'self.errors[0].stacktrace'
        # this allows mutations 'self.stacktrace' to be reflected in the errors
        # list, which is used to generate the JSON payload
        stacktrace = self._generate_stacktrace(
            self.options.pop(
                "traceback",
                getattr(exception, '__traceback__', sys.exc_info()[2])
            ),
            self.options.pop("source_func", None)
        )

        self._stacktrace = stacktrace
        self._errors = self._generate_error_list(exception, stacktrace)

        self.grouping_hash = options.pop("grouping_hash", None)
        self.api_key = options.pop("api_key", get_config("api_key"))

        self.session = None  # type: Optional[Dict]

        self.metadata = {}  # type: Dict[str, Dict[str, Any]]
        if 'meta_data' in options:
            warnings.warn('The Event "metadata" argument has been replaced ' +
                          'with "metadata"', DeprecationWarning)
            for name, tab in options.pop("meta_data").items():
                self.add_tab(name, tab)

        for name, tab in options.pop('metadata', {}).items():
            self.add_tab(name, tab)

        for name, tab in options.items():
            self.add_tab(name, tab)

    @property
    def meta_data(self) -> Dict[str, Dict[str, Any]]:
        warnings.warn('The Event "metadata" property has been replaced ' +
                      'with "meta_data".', DeprecationWarning)
        return self.metadata

    @property
    def breadcrumbs(self) -> List[Breadcrumb]:
        return self._breadcrumbs.copy()

    @property
    def errors(self) -> List[Error]:
        return self._errors.copy()

    @property
    def original_error(self) -> BaseException:
        return self._original_error

    @property
    def stacktrace(self) -> List[Dict[str, Any]]:
        warnings.warn(
            (
                'The Event "stacktrace" property has been deprecated in favour'
                ' of accessing the stacktrace of an error, for example '
                '"errors[0].stacktrace"'
            ),
            DeprecationWarning
        )

        return self._stacktrace

    @stacktrace.setter
    def stacktrace(self, value: List[Dict[str, Any]]) -> None:
        warnings.warn(
            (
                'The Event "stacktrace" property has been deprecated in favour'
                ' of accessing the stacktrace of an error, for example '
                '"errors[0].stacktrace"'
            ),
            DeprecationWarning
        )

        self._stacktrace = value
        self._errors[0].stacktrace = value

    @property
    def exception(self) -> BaseException:
        warnings.warn(
            (
                'The Event "exception" property has been replaced with '
                '"original_error"'
            ),
            DeprecationWarning
        )

        return self._exception

    @exception.setter
    def exception(self, value: BaseException) -> None:
        warnings.warn(
            (
                'Setting the Event "exception" property has been deprecated, '
                'update the "errors" list instead'
            ),
            DeprecationWarning
        )

        self._exception = value
        self._errors[0] = Error(
            class_name(value),
            str(value),
            self._stacktrace
        )

    def set_user(self, id=None, name=None, email=None):
        """
        Set user parameters on event.
        """
        if id:
            self.user["id"] = id
        if name:
            self.user["name"] = name
        if email:
            self.user["email"] = email

    def add_custom_data(self, key, value):
        """
        Add data to the "custom" metadata tab
        """
        self.add_tab("custom", {key: value})

    def add_tab(self, name, dictionary):
        """
        Add a metadata tab to the event

        If the tab already exists, the new content will be merged into the
        existing content.
        """
        if not isinstance(dictionary, dict):
            self.add_tab("custom", {name: dictionary})
            return

        if name not in self.metadata:
            self.metadata[name] = {}

        self.metadata[name].update(dictionary)

    def _generate_error_list(
        self,
        exception: BaseException,
        first_error_stacktrace: List[Dict[str, Any]]
    ) -> List[Error]:
        error_list = [
            Error(
                class_name(exception),
                str(exception),
                first_error_stacktrace
            )
        ]

        if not isinstance(exception, BaseException):
            return error_list

        while True:
            if exception.__cause__:
                exception = exception.__cause__
            elif exception.__context__ and not exception.__suppress_context__:
                exception = exception.__context__
            else:
                break

            error_list.append(
                Error(
                    class_name(exception),
                    str(exception),
                    self._generate_stacktrace(exception.__traceback__)
                )
            )

        return error_list

    def _generate_stacktrace(
        self,
        tb,
        source_func=None
    ) -> List[Dict[str, Any]]:
        """
        Build the stacktrace
        """
        if tb:
            trace = traceback.extract_tb(tb)
        else:
            trace = traceback.extract_stack()

        bugsnag_module_path = os.path.dirname(bugsnag.__file__)
        logging_module_path = os.path.dirname(logging.__file__)
        exclude_module_paths = [bugsnag_module_path, logging_module_path]
        user_exclude_modules = self.config.traceback_exclude_modules
        for exclude_module in user_exclude_modules:
            try:
                module_file = exclude_module.__file__
                if module_file[-4:] == '.pyc':
                    module_file = module_file[:-1]
                exclude_module_paths.append(module_file)
            except Exception:
                self.config.logger.exception(
                    'Could not exclude module: %s' % repr(exclude_module))

        lib_root = self.config.lib_root
        if lib_root and lib_root[-1] != os.sep:
            lib_root += os.sep

        project_root = self.config.project_root
        if project_root and project_root[-1] != os.sep:
            project_root += os.sep

        stacktrace = []
        if source_func is not None:
            try:
                source = inspect.getsourcefile(source_func)
                lines = inspect.getsourcelines(source_func)
                line = 0
                if lines is not None and len(lines) > 1:
                    line = lines[1]

                trace.insert(
                    0,
                    [source, line, source_func.__name__]  # type: ignore
                )
            except (IOError, TypeError):
                pass

        for frame in trace:
            file_name = os.path.abspath(str(frame[0]))
            in_project = False

            skip_module = False
            for module_path in exclude_module_paths:
                if file_name.startswith(module_path):
                    skip_module = True
                    break
            if skip_module:
                continue

            # Fetch the code before (potentially) removing the project root
            # from the file path
            code = self._code_for(file_name, int(str(frame[1])))

            if lib_root and file_name.startswith(lib_root):
                file_name = file_name[len(lib_root):]
            elif project_root and file_name.startswith(project_root):
                file_name = file_name[len(project_root):]
                in_project = True

            stacktrace.append({
                "file": file_name,
                "lineNumber": int(str(frame[1])),
                "method": str(frame[2]),
                "inProject": in_project,
                "code": code
            })

        stacktrace.reverse()

        return stacktrace

    def _code_for(self, file_name, line, window_size=7):
        """
        Find the code around this line in the file.
        """
        if not self.send_code:
            return None

        try:
            lines = linecache.getlines(file_name)

            start = max(line - int(window_size / 2), 1)
            end = start + window_size

            # The last line of the file is len(lines). End of an
            # exclusive range is one greater.
            if end > len(lines) + 1:
                end = len(lines) + 1
                start = max(end - window_size, 1)

            return dict((n, lines[n - 1].rstrip()) for n in range(start, end))

        except Exception:
            return None

    def _payload(self):
        # Fetch the notifier version from the package
        notifier_version = package_version("bugsnag") or "unknown"
        encoder = SanitizingJSONEncoder(
            self.config.logger,
            separators=(',', ':'),
            keyword_filters=self.config.params_filters
        )

        # Construct the payload dictionary
        return encoder.encode({
            "apiKey": self.api_key,
            "notifier": {
                "name": self.NOTIFIER_NAME,
                "url": self.NOTIFIER_URL,
                "version": notifier_version,
            },
            "events": [{
                "severity": self.severity,
                "severityReason": self.severity_reason,
                "unhandled": self.unhandled,
                "releaseStage": self.release_stage,
                "app": {
                    "version": self.app_version,
                    "type": self.app_type,
                },
                "context": self.context,
                "groupingHash": self.grouping_hash,
                "exceptions": [
                    error.to_dict() for error in self.errors
                ],
                "metaData": FilterDict(self.metadata),
                "user": FilterDict(self.user),
                "device": FilterDict({
                    "hostname": self.hostname,
                    "runtimeVersions": self.runtime_versions
                }),
                "projectRoot": self.config.project_root,
                "libRoot": self.config.lib_root,
                "session": self.session,
                "breadcrumbs": [
                    breadcrumb.to_dict() for breadcrumb in self._breadcrumbs
                ]
            }]
        })



from typing import Dict, Any, Tuple, Type
import types
import sys

from bugsnag.breadcrumbs import BreadcrumbType, OnBreadcrumbCallback
from bugsnag.configuration import RequestConfiguration
from bugsnag.client import Client

default_client = Client()
configuration = default_client.configuration
logger = configuration.logger
ExcInfoType = Tuple[Type, Exception, types.TracebackType]


__all__ = ('configure', 'configure_request', 'add_metadata_tab',
           'clear_request_config', 'notify', 'start_session', 'auto_notify',
           'auto_notify_exc_info', 'before_notify', 'leave_breadcrumb')


def configure(**options):
    """
    Configure the Bugsnag notifier application-wide settings.
    """
    return configuration.configure(**options)


def configure_request(**options):
    """
    Configure the Bugsnag notifier per-request settings.
    """
    RequestConfiguration.get_instance().configure(**options)


def add_metadata_tab(tab_name: str, data: Dict[str, Any]):
    """
    Add metaData to the tab

    bugsnag.add_metadata_tab("user", {"id": "1", "name": "Conrad"})
    """
    metadata = RequestConfiguration.get_instance().metadata
    if tab_name not in metadata:
        metadata[tab_name] = {}

    metadata[tab_name].update(data)


def clear_request_config():
    """
    Clears the per-request settings.
    """
    RequestConfiguration.clear()


def notify(exception: BaseException, **options):
    """
    Notify bugsnag of an exception.
    """
    if 'severity' in options:
        options['severity_reason'] = {'type': 'userSpecifiedSeverity'}
    else:
        options['severity_reason'] = {'type': 'handledException'}

    if (isinstance(exception, (list, tuple)) and len(exception) == 3 and
            isinstance(exception[2], types.TracebackType)):
        default_client.notify_exc_info(*exception, **options)
    else:
        if not isinstance(exception, BaseException):
            try:
                value = repr(exception)
            except Exception:
                value = '[BADENCODING]'

            default_client.configuration.logger.warning(
                'Coercing invalid notify() value to RuntimeError: %s' % value
            )

            exception = RuntimeError(value)

        default_client.notify(exception, **options)


def start_session():
    """
    Creates a new session
    """
    default_client.session_tracker.start_session()


def auto_notify(exception: BaseException, **options):
    """
    Notify bugsnag of an exception if auto_notify is enabled.
    """
    if configuration.auto_notify:
        default_client.notify(
            exception,
            unhandled=options.pop('unhandled', True),
            severity=options.pop('severity', 'error'),
            severity_reason=options.pop('severity_reason', {
                'type': 'unhandledException'
            }),
            **options
        )


def auto_notify_exc_info(exc_info: ExcInfoType = None, **options):
    """
    Notify bugsnag of a exc_info tuple if auto_notify is enabled
    """
    if configuration.auto_notify:
        info = exc_info or sys.exc_info()
        if info is not None:
            exc_type, value, tb = info
            default_client.notify_exc_info(
                exc_type, value, tb,
                unhandled=options.pop('unhandled', True),
                severity=options.pop('severity', 'error'),
                severity_reason=options.pop('severity_reason', {
                    'type': 'unhandledException'
                }),
                **options
            )


def before_notify(callback):
    """
    Add a callback to be called before bugsnag is notified

    This can be used to alter the event before sending it to Bugsnag.
    """
    configuration.middleware.before_notify(callback)


def leave_breadcrumb(
    message: str,
    metadata: Dict[str, Any] = {},
    type: BreadcrumbType = BreadcrumbType.MANUAL
) -> None:
    default_client.leave_breadcrumb(message, metadata, type)


def add_on_breadcrumb(on_breadcrumb: OnBreadcrumbCallback) -> None:
    """
    Add a callback to be called each time a breadcrumb is left

    This can be used to alter the breadcrumb, or discard it by returning False
    """
    default_client.add_on_breadcrumb(on_breadcrumb)


def remove_on_breadcrumb(on_breadcrumb: OnBreadcrumbCallback) -> None:
    """
    Remove an existing on_breadcrumb callback
    """
    default_client.remove_on_breadcrumb(on_breadcrumb)


def _auto_leave_breadcrumb(
    message: str,
    metadata: Dict[str, Any],
    type: BreadcrumbType
) -> None:
    default_client._auto_leave_breadcrumb(message, metadata, type)



import os
import platform
import socket
import sys
import sysconfig
from typing import List, Any, Tuple, Union, Optional
import warnings
import logging
from threading import Lock

from bugsnag.breadcrumbs import (
    BreadcrumbType,
    Breadcrumb,
    Breadcrumbs,
    OnBreadcrumbCallback
)
from bugsnag.sessiontracker import SessionMiddleware
from bugsnag.middleware import (
    DefaultMiddleware,
    MiddlewareStack,
    skip_bugsnag_middleware
)
from bugsnag.utils import (fully_qualified_class_name, validate_str_setter,
                           validate_bool_setter, validate_iterable_setter,
                           validate_required_str_setter, validate_int_setter)
from bugsnag.delivery import (create_default_delivery, DEFAULT_ENDPOINT,
                              DEFAULT_SESSIONS_ENDPOINT)
from bugsnag.uwsgi import warn_if_running_uwsgi_without_threads
from bugsnag.error import Error

try:
    from contextvars import ContextVar
    _request_info = ContextVar('bugsnag-request', default=None)  # type: ignore
except ImportError:
    from bugsnag.utils import ThreadContextVar
    _request_info = ThreadContextVar('bugsnag-request', default=None)  # type: ignore  # noqa: E501


__all__ = ('Configuration', 'RequestConfiguration')
_sentinel = object()


class Configuration:
    """
    Global app-level Bugsnag configuration settings.
    """

    def __init__(self, logger=_sentinel):
        self._mutex = Lock()

        self.api_key = os.environ.get('BUGSNAG_API_KEY', None)
        self.release_stage = os.environ.get("BUGSNAG_RELEASE_STAGE",
                                            "production")
        self.notify_release_stages = None
        self.auto_notify = True
        self.send_code = True
        self.send_environment = False
        self.asynchronous = True
        self.delivery = create_default_delivery()
        self.lib_root = sysconfig.get_path('purelib')
        self.project_root = os.getcwd()
        self.app_type = None
        self.app_version = None
        self.params_filters = ["password", "password_confirmation", "cookie",
                               "authorization"]
        self.ignore_classes = [
            "KeyboardInterrupt",
            "django.http.Http404",
            "django.http.response.Http404",
        ]
        self.endpoint = DEFAULT_ENDPOINT
        self.session_endpoint = DEFAULT_SESSIONS_ENDPOINT
        self.auto_capture_sessions = True
        self.traceback_exclude_modules = []

        self.middleware = MiddlewareStack()

        self.internal_middleware = MiddlewareStack()
        self.internal_middleware.before_notify(skip_bugsnag_middleware)
        self.internal_middleware.append(DefaultMiddleware)
        self.internal_middleware.append(SessionMiddleware)

        self.proxy_host = None

        if not os.getenv("DYNO"):
            self.hostname = socket.gethostname()
        else:
            self.hostname = None

        self.runtime_versions = {"python": platform.python_version()}

        self.logger = logger

        self._max_breadcrumbs = 25
        self.breadcrumb_log_level = logging.INFO
        self.enabled_breadcrumb_types = list(BreadcrumbType)
        self._breadcrumbs = Breadcrumbs(self.max_breadcrumbs)
        self._on_breadcrumbs = []

    def configure(self, api_key=None, app_type=None, app_version=None,
                  asynchronous=None, auto_notify=None,
                  auto_capture_sessions=None, delivery=None, endpoint=None,
                  hostname=None, ignore_classes=None, lib_root=None,
                  notify_release_stages=None, params_filters=None,
                  project_root=None, proxy_host=None, release_stage=None,
                  send_code=None, send_environment=None, session_endpoint=None,
                  traceback_exclude_modules=None, logger=_sentinel,
                  breadcrumb_log_level=None, enabled_breadcrumb_types=None,
                  max_breadcrumbs=None):
        """
        Validate and set configuration options. Will warn if an option is of an
        incorrect type.
        """
        if api_key is not None:
            self.api_key = api_key
        if app_type is not None:
            self.app_type = app_type
        if app_version is not None:
            self.app_version = app_version
        if asynchronous is not None:
            self.asynchronous = asynchronous
        if auto_notify is not None:
            self.auto_notify = auto_notify
        if auto_capture_sessions is not None:
            self.auto_capture_sessions = auto_capture_sessions
        if delivery is not None:
            self.delivery = delivery
        if endpoint is not None:
            self.endpoint = endpoint
        if hostname is not None:
            self.hostname = hostname
        if ignore_classes is not None:
            self.ignore_classes = ignore_classes
        if lib_root is not None:
            self.lib_root = lib_root
        if notify_release_stages is not None:
            self.notify_release_stages = notify_release_stages
        if params_filters is not None:
            self.params_filters = params_filters
        if project_root is not None:
            self.project_root = project_root
        if proxy_host is not None:
            self.proxy_host = proxy_host
        if release_stage is not None:
            self.release_stage = release_stage
        if send_code is not None:
            self.send_code = send_code
        if send_environment is not None:
            self.send_environment = send_environment
        if session_endpoint is not None:
            self.session_endpoint = session_endpoint
        if traceback_exclude_modules is not None:
            self.traceback_exclude_modules = traceback_exclude_modules
        if logger is not _sentinel:
            self.logger = logger
        if breadcrumb_log_level is not None:
            self.breadcrumb_log_level = breadcrumb_log_level
        if enabled_breadcrumb_types is not None:
            self.enabled_breadcrumb_types = enabled_breadcrumb_types
        if max_breadcrumbs is not None:
            self.max_breadcrumbs = max_breadcrumbs

        return self

    def get(self, name):
        """
        Get a single configuration option
        """
        warnings.warn('Using get() to retrieve a Configuration property is ' +
                      'deprecated in favor of referencing properties directly',
                      DeprecationWarning)
        return getattr(self, name)

    @property
    def api_key(self):
        """
        Unique application identifier
        """
        return self._api_key

    @api_key.setter  # type: ignore
    @validate_required_str_setter
    def api_key(self, value: str):
        self._api_key = value

    @property
    def app_type(self):
        """
        Category for the current application or task
        """
        return self._app_type

    @app_type.setter  # type: ignore
    @validate_str_setter
    def app_type(self, value: str):
        self._app_type = value

    @property
    def app_version(self):
        """
        Release version of the current application
        """
        return self._app_version

    @app_version.setter  # type: ignore
    @validate_str_setter
    def app_version(self, value: str):
        self._app_version = value

    @property
    def asynchronous(self):
        """
        If API requests should be sent asynchronously
        """
        return self._asynchronous

    @asynchronous.setter  # type: ignore
    @validate_bool_setter
    def asynchronous(self, value: bool):
        self._asynchronous = value
        if value:
            warn_if_running_uwsgi_without_threads()

    @property
    def auto_capture_sessions(self):
        """
        If sessions should be automatically detected and delivered from web
        request integrations
        """
        return self._auto_capture_sessions

    @auto_capture_sessions.setter  # type: ignore
    @validate_bool_setter
    def auto_capture_sessions(self, value: bool):
        self._auto_capture_sessions = value

    @property
    def auto_notify(self):
        """
        If uncaught exceptions should be automatically captured and reported
        """
        return self._auto_notify

    @auto_notify.setter  # type: ignore
    @validate_bool_setter
    def auto_notify(self, value: bool):
        self._auto_notify = value

    @property
    def delivery(self):
        """
        Transport mechanism used to make API requests. Implement the Delivery
        interface to customize how requests are sent.
        """
        return self._delivery

    @delivery.setter  # type: ignore
    def delivery(self, value):
        if hasattr(value, 'deliver') and callable(value.deliver):
            self._delivery = value
        else:
            message = ('delivery should implement Delivery interface, got ' +
                       '{0}. This will be an error in a future release.')
            warnings.warn(message.format(type(value).__name__), RuntimeWarning)

    @property
    def endpoint(self):
        """
        Event API endpoint. Set this property if using Bugsnag On-Premise.

        >>> config = Configuration()
        >>> config.endpoint = 'https://notify.bugsnag.example.co'
        """
        return self._endpoint

    @endpoint.setter  # type: ignore
    @validate_required_str_setter
    def endpoint(self, value: str):
        self._endpoint = value

    @property
    def hostname(self):
        """
        The host name of the application server. This value is automatically
        detected for Heroku applications and included in event device metadata.
        """
        return self._hostname

    @hostname.setter  # type: ignore
    @validate_str_setter
    def hostname(self, value: str):
        self._hostname = value

    @property
    def ignore_classes(self):
        """
        Fully qualified class names which should be ignored when capturing
        uncaught exceptions and other events. KeyboardInterrupt and Http404
        exceptions are ignored by default.
        """
        return self._ignore_classes

    @ignore_classes.setter  # type: ignore
    @validate_iterable_setter
    def ignore_classes(self, value: Union[List[str], Tuple[str]]):
        self._ignore_classes = value

    @property
    def lib_root(self):
        """
        The path to the Python library. Any traceback frame which contains
        lib_root as a prefix is considered out-of-project. The prefix is also
        stripped to make file names easier to read.
        """
        return self._lib_root

    @lib_root.setter  # type: ignore
    @validate_str_setter
    def lib_root(self, value: str):
        self._lib_root = value

    @property
    def notify_release_stages(self):
        """
        A list of release_stage values which are permitted to capture and send
        events and sessions. By default this value is None and all events and
        sessions are delivered.
        """
        return self._notify_release_stages

    @notify_release_stages.setter  # type: ignore
    @validate_iterable_setter
    def notify_release_stages(self, value: List[str]):
        self._notify_release_stages = value

    @property
    def params_filters(self):
        """
        A list of filters applied to event metadata to prevent the values from
        being sent in events. By default the following keys are filtered:

        * authorization
        * cookie
        * password
        * password_confirmation
        """
        return self._params_filters

    @params_filters.setter  # type: ignore
    @validate_iterable_setter
    def params_filters(self, value: List[str]):
        self._params_filters = value

    @property
    def project_root(self):
        """
        The working directory containing the application source code.
        Traceback file paths which contain this prefix are considered a part of
        the project. This prefix is also stripped to increase file name
        readability in traceback lines.
        """
        return self._project_root

    @project_root.setter  # type: ignore
    @validate_str_setter
    def project_root(self, value: str):
        self._project_root = value

    @property
    def proxy_host(self):
        """
        The host name of the proxy to use to deliver requests, if any
        """
        return self._proxy_host

    @proxy_host.setter  # type: ignore
    @validate_str_setter
    def proxy_host(self, value: str):
        self._proxy_host = value

    @property
    def release_stage(self):
        """
        The development phase of the deployed application. This value is used
        to differentiate events which occur in production vs development or
        staging environments.
        """
        return self._release_stage

    @release_stage.setter  # type: ignore
    @validate_str_setter
    def release_stage(self, value: str):
        self._release_stage = value

    @property
    def send_code(self):
        """
        If the source code lines immediately surrounding traceback locations
        should be sent with events
        """
        return self._send_code

    @send_code.setter  # type: ignore
    @validate_bool_setter
    def send_code(self, value: bool):
        self._send_code = value

    @property
    def send_environment(self):
        """
        If the request environment should be automatically collected and
        attached to events
        """
        return self._send_environment

    @send_environment.setter  # type: ignore
    @validate_bool_setter
    def send_environment(self, value: bool):
        self._send_environment = value

    @property
    def session_endpoint(self):
        """
        Sessions API endpoint. Set this property if using Bugsnag On-Premise.

        >>> config = Configuration()
        >>> config.session_endpoint = 'https://sessions.bugsnag.example.co'
        """
        return self._session_endpoint

    @session_endpoint.setter  # type: ignore
    @validate_required_str_setter
    def session_endpoint(self, value: str):
        self._session_endpoint = value

    @property
    def traceback_exclude_modules(self):
        """
        Modules which should be stripped from event tracebacks entirely
        """
        return self._traceback_exclude_modules

    @traceback_exclude_modules.setter  # type: ignore
    @validate_iterable_setter
    def traceback_exclude_modules(self, value: List[str]):
        self._traceback_exclude_modules = value

    @property
    def logger(self) -> logging.Logger:
        """
        Logger for use internally
        """
        return self._logger

    @logger.setter  # type: ignore
    def logger(self, logger: Optional[logging.Logger]) -> None:
        if logger is _sentinel:
            logger = self._create_default_logger()
        elif logger is None:
            logger = self._create_null_logger()

        if not isinstance(logger, logging.Logger):
            actual = type(logger).__name__
            message = 'logger should be logging.Logger, got ' + actual
            warnings.warn(message, RuntimeWarning)

            logger = self._create_default_logger()

        self._logger = logger

    @property
    def max_breadcrumbs(self) -> int:
        return self._max_breadcrumbs

    @max_breadcrumbs.setter  # type: ignore
    @validate_int_setter
    def max_breadcrumbs(self, new_max: int) -> None:
        if 0 <= new_max <= 100:
            self._breadcrumbs.resize(new_max)
            self._max_breadcrumbs = new_max
        else:
            message = (
                'max_breadcrumbs should be an int between 0 and 100, got "{}"'
            ).format(new_max)

            warnings.warn(message, RuntimeWarning)

    @property
    def breadcrumbs(self) -> List[Breadcrumb]:
        return self._breadcrumbs.to_list()

    def add_on_breadcrumb(self, on_breadcrumb: OnBreadcrumbCallback) -> None:
        with self._mutex:
            self._on_breadcrumbs.append(on_breadcrumb)

    def remove_on_breadcrumb(
        self,
        on_breadcrumb: OnBreadcrumbCallback
    ) -> None:
        with self._mutex:
            try:
                self._on_breadcrumbs.remove(on_breadcrumb)
            except ValueError:
                # ignore exception if "on_breadcrumb" is not in the list
                pass

    def should_notify(self) -> bool:
        return self.notify_release_stages is None or \
            (isinstance(self.notify_release_stages, (tuple, list)) and
             self.release_stage in self.notify_release_stages)

    def should_ignore(
        self,
        exception: Union[BaseException, List[Error]]
    ) -> bool:
        if self.ignore_classes is None:
            return False

        # if this is a list of Error instances, check if any of the error
        # classes are ignored
        if isinstance(exception, list):
            return any(e.error_class in self.ignore_classes for e in exception)

        return fully_qualified_class_name(exception) in self.ignore_classes

    def _create_default_logger(self) -> logging.Logger:
        logger = logging.getLogger('bugsnag')
        logger.setLevel(logging.WARNING)

        format = '%(asctime)s - [%(name)s] %(levelname)s - %(message)s'
        formatter = logging.Formatter(format)

        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(formatter)

        logger.addHandler(handler)

        return logger

    def _create_null_logger(self) -> logging.Logger:
        logger = logging.getLogger('bugsnag')
        logger.handlers = []

        logger.addHandler(logging.NullHandler())

        return logger


class RequestConfiguration:
    """
    Per-request Bugsnag configuration settings.
    """

    @classmethod
    def get_instance(cls):
        """
        Get this thread's instance of the RequestConfiguration.
        """

        try:
            instance = _request_info.get()
        except LookupError:
            instance = None

        if instance is None:
            instance = RequestConfiguration()
            _request_info.set(instance)  # type: ignore

        return instance

    @classmethod
    def clear(cls):
        """
        Clear this thread's instance of the RequestConfiguration.
        """
        _request_info.set(None)

    def __init__(self):
        self.context = None
        self.grouping_hash = None
        self.user = {}
        self.metadata = {}

        # legacy fields
        self.user_id = None
        self.extra_data = {}
        self.request_data = {}
        self.environment_data = {}
        self.session_data = {}

    def get(self, name) -> Any:
        """
        Get a single configuration option
        """
        return getattr(self, name)

    def configure(self, **options):
        """
        Set one or more configuration settings.
        """
        for name, value in options.items():
            setattr(self, name, value)

        return self

    @property
    def meta_data(self) -> Any:
        warnings.warn('RequestConfiguration.meta_data has been renamed to ' +
                      '"metadata"', DeprecationWarning)
        return self.metadata



from webob import Request
import sys
from typing import Dict

import bugsnag
from bugsnag.wsgi import request_path
from bugsnag.breadcrumbs import BreadcrumbType
from bugsnag.legacy import _auto_leave_breadcrumb
from bugsnag.utils import remove_query_from_url

# Attempt to import bottle for runtime version report, but only if already
# in use in app
bottle_present = False
if 'bottle' in sys.modules:
    bottle_present = True
    try:
        import bottle
    except ImportError:
        bottle_present = False


__all__ = ('BugsnagMiddleware',)


def add_wsgi_request_data_to_notification(event):
    if not hasattr(event.request_config, "wsgi_environ"):
        return

    environ = event.request_config.wsgi_environ
    request = Request(environ)
    event.request = request
    path = request_path(environ)

    event.context = "%s %s" % (request.method, path)
    event.set_user(id=request.client_addr)
    event.add_tab("request", {
        "url": "%s%s" % (request.application_url, path),
        "headers": dict(request.headers),
        "params": dict(request.params),
    })

    if bugsnag.configure().send_environment:
        event.add_tab("environment", dict(request.environ))


class WrappedWSGIApp:
    """
    Wraps a running WSGI app and sends all exceptions to bugsnag.
    """

    SEVERITY_REASON = {
        "type": "unhandledExceptionMiddleware",
        "attributes": {
            "framework": "WSGI"
        }
    }

    def __init__(self, application, environ, start_response):
        self.environ = environ

        bugsnag.configure_request(wsgi_environ=self.environ)
        try:
            if bugsnag.configuration.auto_capture_sessions:
                bugsnag.start_session()

            _auto_leave_breadcrumb(
                'http request',
                _get_breadcrumb_metadata(environ),
                BreadcrumbType.NAVIGATION
            )

            self.app = application(environ, start_response)
        except Exception as e:
            bugsnag.auto_notify(
                e,
                severity_reason=self.SEVERITY_REASON
            )
            raise

    def __iter__(self):
        try:
            for response in self.app:
                yield response

        except Exception as e:
            bugsnag.auto_notify(
                e,
                severity_reason=self.SEVERITY_REASON
            )
            raise

    def close(self):
        try:
            if hasattr(self.app, 'close'):
                return self.app.close()
        except Exception as e:
            bugsnag.auto_notify(
                e,
                severity_reason=self.SEVERITY_REASON
            )
            raise
        finally:
            bugsnag.clear_request_config()


def _get_breadcrumb_metadata(environ) -> Dict[str, str]:
    metadata = {}

    if 'PATH_INFO' in environ:
        metadata['to'] = environ['PATH_INFO']

    if 'HTTP_REFERER' in environ:
        metadata['from'] = remove_query_from_url(environ['HTTP_REFERER'])

    return metadata


class BugsnagMiddleware:
    """
    Notifies Bugsnag on any unhandled exception that happens while processing
    a request in your application.

    This middleware should be installed before any middlewares that catch
    the exception and render an error messae.
    """

    def __init__(self, application):
        middleware = bugsnag.configure().internal_middleware
        if bottle_present:
            bugsnag.configure().runtime_versions['bottle'] = bottle.__version__
        middleware.before_notify(add_wsgi_request_data_to_notification)
        self.application = application

    def __call__(self, environ, start_response):
        return WrappedWSGIApp(self.application, environ, start_response)



try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    from bugsnag.django.utils import MiddlewareMixin

from typing import Dict

import bugsnag
import bugsnag.django
from bugsnag.legacy import _auto_leave_breadcrumb
from bugsnag.breadcrumbs import BreadcrumbType
from bugsnag.utils import remove_query_from_url


class BugsnagMiddleware(MiddlewareMixin):
    def __init__(self, get_response=None):
        self.config = bugsnag.django.configure()
        super(BugsnagMiddleware, self).__init__(get_response)

    # pylint: disable-msg=R0201
    def process_request(self, request):
        bugsnag.configure_request(django_request=request)

        _auto_leave_breadcrumb(
            'http request',
            self._get_breadcrumb_metadata(request),
            BreadcrumbType.NAVIGATION
        )

        return None

    # pylint: disable-msg=W0613
    def process_response(self, request, response):
        self.config._breadcrumbs.clear()
        bugsnag.clear_request_config()

        return response

    def process_exception(self, request, exception):
        try:
            bugsnag.auto_notify(
                exception,
                severity_reason={
                    "type": "unhandledExceptionMiddleware",
                    "attributes": {
                        "framework": "Django"
                    }
                }
            )

            # ensure the 'got_request_exception' signal doesn't also notify
            # this exception
            setattr(exception, 'skip_bugsnag', True)

        except Exception:
            self.config.logger.exception("Error in exception middleware")

        bugsnag.clear_request_config()

        return None

    def _get_breadcrumb_metadata(self, request) -> Dict[str, str]:
        metadata = {'to': request.path}

        if 'HTTP_REFERER' in request.META:
            metadata['from'] = remove_query_from_url(
                request.META['HTTP_REFERER']
            )

        return metadata



import warnings

from bugsnag.event import Event


__all__ = ('Notification',)


class Notification(Event):
    def __init__(self, exception, config, request_config, **options):
        warnings.warn('The Notification class has been deprecated in favor ' +
                      'of bugsnag.event.Event and will be removed in a ' +
                      'future release.', DeprecationWarning)
        super().__init__(exception, config, request_config, **options)



from typing import Any, Dict, List

__all__ = ('Error',)


class Error:
    __slots__ = (
        'error_class',
        'error_message',
        'stacktrace',
        'type',
    )

    def __init__(
        self,
        error_class: str,
        error_message: str,
        stacktrace: List[Dict[str, Any]]
    ):
        self.error_class = error_class
        self.error_message = error_message
        self.stacktrace = stacktrace
        self.type = 'python'

    def to_dict(self) -> Dict[str, Any]:
        return {
            'errorClass': self.error_class,
            'message': self.error_message,
            'stacktrace': self.stacktrace,
            'type': self.type,
        }
